from __future__ import annotations
from typing import Type, Protocol
from abc import ABC, abstractmethod
from enum import Flag, auto
import typing
from type_enforcer import TypeEnforcer

# JSON serializable primitive types
type JsonPrimitiveTypes = str | int | float | bool | dict | list | None

# This is a type alias for all types that can be serialized to JSON
type JsonSerializableTypes = (
    JsonPrimitiveTypes | dict[typing.Any, typing.Any] | list[typing.Any]
)


class SettingValue(ABC):
    @abstractmethod
    def to_dict(self) -> dict:
        pass

    @abstractmethod
    def from_dict(self, dict_: dict):
        pass


class SettingValueProtocol(Protocol):
    def to_dict(self) -> dict: ...

    @abstractmethod
    def from_dict(self, dict_: dict): ...



class Level(Flag):
    USER = auto()
    USER_DEV = auto()   # For users with dev_mode
    DEVELOPER = auto()  # For plugin developers
    ADVANCED = auto()
    READ_ONLY = auto()
    HIDDEN = auto()
    
class SettingType(Flag):
    PERFORMANCE = auto()
    COSMETIC = auto()
    QOL = auto()
    OTHER = auto()

class Setting[T: SettingValue | SettingValueProtocol | JsonSerializableTypes](
    TypeEnforcer[T]
):
    JsonSerializableTypes = JsonSerializableTypes
    SettingValue = SettingValue
    SettingValueProtocol = SettingValueProtocol
    
    Level = Level
    SettingType = SettingType

    def __init__(
        self,
        value: T,
        name: str = "",
        unit: str = "",     # TODO: Better units
        options=None,
        level: Level=Level.USER,
        type_: SettingType=SettingType.OTHER,
        description="",
        strongly_typed: bool = True,
    ):
        """Class to create setting in the Config. Is strongly typed.
        Create with `Setting[value_type](value)`.

        Value type must either be `Setting.SettingValue`, of `Setting.JsonSerializableTypes` type or use `Setting.SettingValueProtocol`

        By default Setting will not try parsing `value` into instance of `value_type`. Set `strongly_typed` to `False` to change it.

        Value of the class can be accessed by calling the instance:

        ```number = Setting[int](15, 'Some Number')
        number()   # returns a number._value (15)
        ```
        """
        super().__init__(value, strongly_typed)
        self.name = name
        self.unit = unit
        self.options = options
        self.level = level
        self.type_ = type_
        self.description = description

        self._type: Type[T] | None = None

    def __call__(self) -> T:
        return self._value

    def __set__(self, instance, value: T) -> None:
        self._check_type(value, self._type)

    def __str__(self) -> str:
        return f"{self.name}: {self._type.__name__}({self()}{f' {self.unit}' if self.unit else ''})"    # type: ignore

    def to_dict(self) -> dict[str, JsonSerializableTypes]:
        """Convert setting to dictionary."""
        result = {
            "name": self.name,
            "options": self.options,
            "description": self.description,
            "type": self.type_,
        }
        if self._check_type(self._type, JsonSerializableTypes):
            result["value"] = self._value
        else:
            result["value"] = self._value.to_dict()
        
        return result


from __future__ import annotations
from pathlib import Path
import typing
import json
from setting import Setting


class ConfigMeta(type):
    """Metaclass for Config to enable proper inheritance and access patterns."""

    def __new__(mcs, name: str, bases: tuple, namespace: dict) -> typing.Type:
        cls = super().__new__(mcs, name, bases, namespace)
        setattr(cls, '_config_name', name)

        # Process nested Config classes to make them proper attributes
        for key, value in list(cls.__dict__.items()):
            if (
                isinstance(value, type)
                and issubclass(value, Config)
                and value is not Config
            ):
                # Replace the class reference with an instance
                setattr(cls, key, value())

        return cls


class Config(metaclass=ConfigMeta):
    """Base configuration class."""

    def to_dict(self) -> dict[str, typing.Any]:
        """Convert config to dictionary recursively."""
        result = {}

        # Process all attributes including inherited ones
        for key, value in vars(self.__class__).items():
            if key.startswith("_"):
                continue

            if isinstance(value, Setting):
                if hasattr(value(), "to_dict") and callable(value().to_dict):
                    result[key] = value().to_dict()
                else:
                    result[key] = value()
            elif isinstance(value, Config):
                result[key] = value.to_dict()

        return result

    # def to_dict_with_metadata(self) -> dict[str, ]:   # TODO: May be

    @classmethod
    def from_dict(cls, data: dict[str, typing.Any]) -> Config:
        """Create config instance from dictionary."""
        instance = cls()

        for key, value in data.items():
            attr = getattr(instance.__class__, key, None)

            if isinstance(attr, Setting):
                setattr(instance.__class__, key, value)
            elif isinstance(attr, Config) and isinstance(value, dict):
                # Recursively set nested config
                attr.__class__.from_dict(value)

        return instance

    def save(self, file_path: str | Path) -> None:
        """Save config to JSON file."""
        with open(file_path, "w") as f:
            json.dump(self.to_dict(), f, indent=4)

    @classmethod
    def load(cls, file_path: str | Path) -> Config:
        """Load config from JSON file."""
        with open(file_path, "r") as f:
            data = json.load(f)
        return cls.from_dict(data)

    def get_settings_dict(self) -> dict[str | Config, Setting]:
        result = {}

        for key, value in vars(self.__class__).items():
            if key.startswith("_"):
                continue

            if isinstance(value, Setting):
                result[key] = value
            elif isinstance(value, Config):
                result[key] = value.get_settings_dict()

        return result

    def get_all_settings(self) -> dict[str, Setting]:
        """Get all settings recursively with their full paths."""
        result = {}

        def collect_settings(config, prefix=""):
            for key, value in vars(config.__class__).items():
                if key.startswith("_"):
                    continue

                full_path = f"{prefix}{key}" if prefix else key

                if isinstance(value, Setting):
                    result[full_path] = value
                elif isinstance(value, Config):
                    collect_settings(value, f"{full_path}.")

        collect_settings(self)
        return result

    def __str__(self) -> str:
        def handle_dict(dict_: dict[str, Setting | dict], indent: int = 1) -> str:
            s = ""
            for key, value in dict_.items():
                if isinstance(value, dict):
                    s += "  " * indent + f"{key}:\n"
                    s += handle_dict(value, indent + 1)
                else:
                    s += "  " * indent + f"{value._attribute_name}: {str(value)}\n"
            return s

        return f"{self._config_name}:\n" + handle_dict(self.get_settings_dict())


import multiprocessing
from config import Config
from setting import Setting, Level, SettingType


class AppConfig(Config):
    version = Setting[str]('0.1.0', 'Version', level=Level.USER | Level.READ_ONLY)
    dev_mode = Setting[bool](True, 'Dev Mode', level=Level.USER)
    debug_mode = Setting[bool](True, 'Debug Mode', level=Level.USER_DEV)
    
    class ImageDownloading(Config):
        convert_image = Setting[bool](True, 'Convert Image')
        preferable_format = Setting[str]('WEBP', 'Converted Images Format')
        
        max_threads = Setting[int](multiprocessing.cpu_count(), 'Max Download Threads')
        image_update_every = Setting[int](10, 'Image Update Percentage', '%')
        
        
        PIL_SUPPORTED_EXT = Setting[dict[str, str]]({
            'JPG':  'JPEG',
            'JPEG': 'JPEG',
            'PNG':  'PNG',
            'GIF':  'GIF',
            'BMP':  'BMP',
            'WEBP': 'WEBP',
            'ICO':  'ICO',
            'TIFF': 'TIFF',
            'TIF':  'TIFF'
        }, 'Formats that PIL supports')
        
    class UI(Config):
        class MangaViewer(Config):
            image_loading_intervals = Setting[int](100, 'Load Images in UI with Intervals', 'ms')
            placeholder_loading_intervals = Setting[int](5, 'Load Placeholders in UI with Intervals', 'ms')
            
            set_size_with_every_placeholder = Setting[bool](True, 'Set MangaViewer\'s Scene Size with Every Placeholder Added', level=Level.USER | Level.ADVANCED)
            cull_height_multiplier = Setting[float](2.0, 'Cull Viewport Height Multiplier', level=Level.USER | Level.ADVANCED, type_=SettingType.PERFORMANCE | SettingType.COSMETIC)
            cull_scene_cooldown = Setting[int](250, 'Scene Culling Minimum Cooldown', 'ms', level=Level.USER | Level.ADVANCED, type_=SettingType.PERFORMANCE)
        
    class Scrolling(Config):
        step = Setting[int](150, 'Step', 'px', type_=SettingType.COSMETIC | SettingType.QOL)
        step_duration = Setting[int](200, 'Step Duration', 'ms', type_=SettingType.COSMETIC | SettingType.QOL)
        alt_multiplier = Setting[int](8, 'Alt Step Multiplier', type_=SettingType.QOL)
        
        scale_multiplier = Setting[float](1.5, 'Step Scale Multiplier', level=Level.DEVELOPER, type_=SettingType.PERFORMANCE)
        
        
print(AppConfig())


from __future__ import annotations
from typing import (
    Any,
    Type,
    Sequence,
    Set,
    MutableSet,
    MutableSequence,
    Mapping,
    MutableMapping,
    get_type_hints,
    get_origin,
    get_args,
)
from types import UnionType


class TypeEnforcer[T]:
    """Class will enforce `T` type on `value`"""

    def __init__(
        self, value: T, strongly_typed: bool = True, try_parse_after_failure=False
    ):
        """
        Args:
            value (T): value to be enforced
            strongly_typed (bool, optional): If set to `True`, will immediately raise a TypeError if value's type is wrong.
            Otherwise, will try to parse the value by invoking T(value). Defaults to True.
            try_parse_after_failure (bool, optional): Will try to parse all values even if parsing of one failed. Defaults to False.
        """
        self._value = value
        self.strongly_typed = strongly_typed
        self.try_parse_after_error = try_parse_after_failure

    def __set_name__(self, owner: Type, name: str):
        """Called when the TypeEnforcer is declared in a class."""
        self._attribute_owner = owner
        self._attribute_name = name

        if hasattr(self, "__orig_class__"):
            origin = get_origin(self.__orig_class__)    # type: ignore
            args = get_args(self.__orig_class__)        # type: ignore
        elif name in (hints := get_type_hints(owner)):
            origin = get_origin(hints[name])
            args = get_args(origin)
        else:
            raise TypeError("Type must be set, either in type hint, or in generic type")

        if origin and args:
            self._type = args[0]

        if self.strongly_typed or isinstance(self._value, (tuple, frozenset)):
            parsed = self._check_type(self._value, self._type)
            val = None
        else:
            parsed, val = self._parse_check_type(
                self._value, self._type, self.try_parse_after_error
            )

        if parsed:
            if val is not None:
                self._value = val
        else:
            if self.strongly_typed:
                raise TypeError(
                    f"Value of class {owner}.{name} must be of type {self._type}"
                )
            else:
                raise TypeError(
                    f"Value of class {owner}.{name} must be of type {self._type} or corresponding values must be able to convert into corresponding types.\n"
                    f"Unable to convert {self._value} in type {self._type}"
                )

    @classmethod
    def _check_type(cls, value, type_) -> bool:
        """
        Checks if the value is of the given type.

        Args:
            `value`: The value to check.
            `type_`: The type to check against.

        Returns:
            True if the value is of the given type, False otherwise.
        """
        orig = get_origin(type_)
        args = get_args(type_)

        # Is simple type (aka int, str)
        if not orig:
            return isinstance(value, type_)

        # Is union (aka int | str)
        elif orig is UnionType:
            return any(cls._check_type(value, arg) for arg in args)

        # Is complex type (aka dict[str, int], list[dict[str, str | int]])
        else:
            if isinstance(value, orig):
                if isinstance(value, list) or issubclass(
                    orig, (Sequence, MutableSequence)
                ):
                    return all(map(lambda x: cls._check_type(x, args[0]), value))

                if isinstance(value, dict) or issubclass(
                    orig, (Mapping, MutableMapping)
                ):
                    return all(
                        map(lambda x: cls._check_type(x, args[0]), value.keys())
                    ) and all(
                        map(lambda x: cls._check_type(x, args[1]), value.values())
                    )

            else:
                return False

        raise TypeError(f"Type '{type_}' is not supported yet")

    @classmethod
    def _parse_check_type(
        cls, value, type_, ignore_error=False
    ) -> tuple[bool, T | Any]:
        """
        Attempts to parse a value to a specified type, supporting complex and union types.

        This method checks whether a given value can be parsed into the specified type.
        If the value is not of the specified type, it attempts to convert it. The method
        supports simple types, union types, sequences, set, and mappings. If the conversion
        is successful, it returns the parsed value; otherwise, it returns the failed value,
        see `ignore_error`.

        Args:
            `value`: The value to be checked and potentially parsed.
            `type_`: The target type to parse the value into.
            `ignore_error` (bool, optional): If set to True, the method will attempt to parse
                                        all elements in sequences and mappings even if
                                        parsing of one element fails. Defaults to False.

        Returns:
            A tuple where the first element is a boolean indicating if the parsing was
            successful, and the second element is the parsed value or the original value
            if parsing was unsuccessful.
        """

        orig = get_origin(type_)
        args = get_args(type_)

        # Is simple type (aka int, str)
        if not orig:
            if not isinstance(value, type_):
                try:
                    return True, type_(value)
                except (ValueError, TypeError):
                    return False, value
            return True, value

        # Is union (aka int | str)
        elif orig is UnionType:
            for arg in args:
                parsed, v = cls._parse_check_type(value, arg, ignore_error)
                if parsed:
                    break
            return parsed, v

        # Is complex type (aka dict[str, int], list[dict[str, str | int]])
        else:
            if isinstance(value, orig):
                if is_list := isinstance(value, list) or issubclass(
                    orig, (Sequence, MutableSequence, Set, MutableSet)
                ):
                    result_list = []
                    parsed = True
                    for v in value:
                        values_parsed, val = cls._parse_check_type(
                            v, args[0], ignore_error
                        )
                        result_list.append(val)
                        if not values_parsed:
                            parsed = False
                            if not ignore_error:
                                break
                    return parsed, result_list if is_list else type(value)(result_list)

                if isinstance(value, dict) or issubclass(
                    orig, (Mapping, MutableMapping)
                ):
                    result_dict = {}
                    parsed = True
                    for k, v in value.items():
                        keys_parsed, key = cls._parse_check_type(
                            k, args[0], ignore_error
                        )
                        values_parsed, val = cls._parse_check_type(
                            v, args[1], ignore_error
                        )
                        result_dict[key] = val
                        if not keys_parsed or not values_parsed:
                            parsed = False
                            if not ignore_error:
                                break
                    return parsed, result_dict

            else:
                return False, value
